#!/usr/bin/env python
"""Collect command-line options in a dictionary"""

import pandas as pd
import logging
import os

from oplrareg import *
from sklearn import preprocessing

logger = logging.getLogger(__name__)


def getopts(argv):
    opts = {}  # Empty dictionary to store key-value pairs.
    while argv:  # While there are arguments left to parse...
        if argv[0][0] == "-" or argv[0][0] == "--":  # Found a "-name value" pair.
            opts[argv[0]] = argv[1]  # Add key and value to the dictionary.
        argv = argv[1:]  # Reduce the argument list by copying it starting from index 1.
    return opts


def parse_input(input_file):
    if input_file.endswith(".tab") or input_file.endswith(".data"):
        logger.info("Reading tabular data")
        contents = pd.read_csv(input_file, skiprows=1, sep=r"\s+")
    elif input_file.endswith(".csv"):
        logger.info("Reading .csv file")
        contents = pd.read_csv(input_file)
    elif input_file.endswith(".xls") or input_file.endswith(".xlsx"):
        logger.info("Reading spreadsheet")
        contents = pd.read_excel(input_file)
        raise ValueError("Input file extension is not recognised:  %s" % input_file)
    X = contents.iloc[:, :-1]
    y = contents.iloc[:, -1]

    return X, y


def main():
    from sys import argv

    myargs = getopts(argv)
    if "--input" in myargs or "-i" in myargs:
        if "--input" in myargs:
            input_file = myargs["--input"]
        else:
            input_file = myargs["-i"]
        X, y = parse_input(input_file)
    else:
        msg = (
            "Please inform an input file with --input or -i argument.\n"
            + "\tExample: \n\t\toplrareg --input yacht_hydrodynamics.data\n"
            + "\t\toplrareg -i yacht_hydrodynamics.data"
        )
        raise ValueError(msg)

    min_max_scaler = preprocessing.MinMaxScaler()
    import ipdb; ipdb.set_trace()
    scaled_X = pd.DataFrame(min_max_scaler.fit_transform(X), columns=X.columns)

    params = {}
    fit_params = {}

    if "-l" in myargs:  # Example usage.
        params["lam"] = float(myargs["-l"])
    elif "--lambda" in myargs:
        params["lam"] = float(myargs["--lambda"])

    if "-e" in myargs:  # Example usage.
        params["epsilon"] = float(myargs["-e"])
    elif "--epsilon" in myargs:
        params["epsilon"] = float(myargs["--epsilon"])

    if "-b" in myargs:  # Example usage.
        params["beta"] = float(myargs["-b"])
    elif "--beta" in myargs:
        params["beta"] = float(myargs["--beta"])

    if "-s" in myargs:  # Example usage.
        params["solver_name"] = myargs["-b"]
    elif "--solver" in myargs:
        params["solver_name"] = myargs["--solver"]

    if "-r" in myargs:  # Example usage.
        params["exact_number_regions"] = int(myargs["-r"])
    elif "--regions" in myargs:
        params["exact_number_regions"] = int(myargs["--regions"])

    if "-pf" in myargs:  # Example usage.
        fit_params["f_star"] = myargs["-f"]
    elif "--partition-feature" in myargs:
        fit_params["f_star"] = myargs["--partition-feature"]

    oplra_reg = OplraRegularised()
    if len(params) > 0:
        oplra_reg.set_params(**params)

    if len(fit_params) > 0:
        oplra_reg.fit(scaled_X, y, f_star=fit_params["f_star"])
    else:
        oplra_reg.fit(scaled_X, y)

    coeffs, breakpoints = oplra_reg.get_model_info()

    print("BREAKPOINTS:")
    print(breakpoints)
    print()

    print("COEFFICIENTS:")
    print(coeffs)
    print()

    print("MEAN ABSOLUTE ERROR: %f" % oplra_reg.final_model.mae.value)

    input_name = os.path.splitext(input_file)[0]

    coeffs.to_csv(input_name + "_coefficients.csv", index=False)
    breakpoints.to_csv(input_name + "_breakpoints.csv", index=False)


if __name__ == "__main__":
    main()



from pyomo.environ import SolverFactory
from abc import ABCMeta, abstractmethod


def get_solver_definition(solver_name, **kwargs):
    if solver_name.lower() == "cplex":
        return CplexDefinition(**kwargs)
    elif solver_name.lower() == "glpk":
        return GlpkDefinition(**kwargs)
    else:
        raise ValueError('Solver %s is not supported' % solver_name)


class SolverDefinition(metaclass=ABCMeta):
    def __init__(self, solver_name, time_limit=120, threads=4, mipgap=0.01):
        self.timeLimit = time_limit
        self.threads = threads
        self.mipgap = mipgap
        self.solver = SolverFactory(solver_name)
        self.update()

    @abstractmethod
    def update(self):
        pass

    def solve(self, pyomo_model, verbose=False):
        if verbose:
            results = self.solver.solve(pyomo_model, keepfiles=True,
                                        tee=True, symbolic_solver_labels=True)
        else:
            results = self.solver.solve(pyomo_model, keepfiles=False,
                                        tee=False, symbolic_solver_labels=True)
        pyomo_model.solutions.load_from(results)
        return results.solver.termination_condition


class CplexDefinition(SolverDefinition):
    def __init__(self, **kwargs):
        super().__init__("cplex", **kwargs)

    def update(self):
        self.solver.options['timelimit'] = self.timeLimit
        self.solver.options['threads'] = self.threads
        self.solver.options['mipgap'] = self.mipgap


class GlpkDefinition(SolverDefinition):
    def __init__(self, **kwargs):
        super().__init__("glpk", **kwargs)

    def update(self):
        self.solver.options['tmlim'] = self.timeLimit
        self.solver.options['mipgap'] = self.mipgap



import pandas as pd

from collections import Counter
from pyomo.opt import TerminationCondition

from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import safe_sparse_dot

from .models import *
from .base import *


class OplraRegularised(BaseOplraEstimator):
    """Runs Oplra with regularisation

    Attributes:
        lam          (float) : Regularisation parameter
        epsilon      (float) : MIP parameter, small number to separate the breakpoints
        beta         (float) : Stopping criteria, usually set to 0.03. Algorithm will stop when the improvement
                                in the objective function is no more than beta.
        exact_number_regions (int) : If specified, runs Oplra to the the number of regions passed to this argument
        solver_name     (str) : Name of a supported solver (CPLEX or GLPK)
    """

    def __init__(
        self,
        lam=0.005,
        epsilon=0.01,
        beta=0.03,
        exact_number_regions=None,
        solver_name="glpk",
    ):
        super().__init__("OplraRegularised", "v1", lam, epsilon, beta, solver_name)
        self.final_model = None
        if exact_number_regions is None:
            self.exact_number_regions = None
            self.beta = beta
        else:
            self.beta = None
            self.exact_number_regions = exact_number_regions

    def predict(self, X):
        """Predicts outcome of new data

        :param X: data to predict outcome
        :return: list with predictions for each sample in X
        """

        # It is necessary to run `fit` function before `predict`
        check_is_fitted(self, "final_model")

        samples = range(X.shape[0])

        # Defines the regions of the samples
        if self.final_model.number_regions > 1:
            if type(X) == pd.DataFrame:
                ts_region = self.final_model.get_regions(X.loc[:, self.final_model.fStar])
            else:
                # Then it is a numpy array

                fStar_idx = None
                for idx, f in enumerate(self.final_model.f):
                    if f == self.final_model.fStar:
                        fStar_idx = idx
                        break

                ts_region = self.final_model.get_regions(X[:, fStar_idx])

        else:
            ts_region = np.zeros(len(samples))

        counter_region = Counter(ts_region)
        final_predictions = np.zeros(len(samples))
        coefficients = self.final_model.get_coefficients()
        intercepts = self.final_model.get_intercepts()

        if len(counter_region) == 1:

            which_region = counter_region.most_common(1)[0][0]

            final_predictions = (
                safe_sparse_dot(X, coefficients[which_region], dense_output=True) + intercepts[which_region]
            )
        else:
            for region, counter_region in counter_region.items():
                samples_in_region = np.argwhere(ts_region == region).transpose()[0]
                if type(X) == pd.DataFrame:
                    X_samples = X.iloc[samples_in_region]
                else:
                    X_samples = X[samples_in_region, ]

                final_predictions[samples_in_region] = (
                    safe_sparse_dot(
                        X_samples,
                        coefficients[region],
                        dense_output=True,
                    )
                    + intercepts[region]
                )

        return final_predictions

    def fit(self, X, y, f_star=None, verbose=True):
        """

        :param X:
        :param y:
        :param f_star
        :param verbose
        :return:
        """

        if verbose:
            print("%s\n########## R = 1" % self)

        simple_linear_model, termination_condition = self.run_model(X, y, 1)
        selected_features = simple_linear_model.get_selected_features()

        print("SELECTED FEATURES:")
        print(selected_features)
        print()

        if len(selected_features) == 0:
            self.final_model = simple_linear_model
            return self

        # Solve for exact number of regions == 1 or when exact number of regions >= 2 when f* was defined beforehand
        if self.beta is None or self.exact_number_regions == 1:
            self.final_model = simple_linear_model
            return self
        elif f_star is not None:
            number_regions = self.exact_number_regions
            tmp_model, termination_condition = self.run_model(
                X, y, number_regions, f_star, selected_features
            )
            self.final_model = tmp_model
            return self

        # ---- PIECEWISE MODEL R = 2 ---- #
        number_regions = 2
        best_model = None
        count = 1
        print(f_star)

        if f_star is None:
            # ---- Find best f* for 2 regions ---- #
            for f_star in selected_features:
                if verbose:
                    print(
                        "\n%s\n########## R = %d f* = %s (loop %d/%d) ###########"
                        % (self, number_regions, f_star, count, len(selected_features))
                    )

                tmp_model, termination_condition = self.run_model(
                    X, y, number_regions, f_star, selected_features
                )
                count += 1
                if termination_condition != TerminationCondition.infeasible and (
                    best_model is None or tmp_model.z.value < best_model.z.value
                ):
                    best_model = tmp_model
        else:
            if verbose:
                print(
                    "\n%s\n########## R = %d f* = %s ###########"
                    % (self, number_regions, f_star)
                )

            tmp_model, termination_condition = self.run_model(
                X, y, number_regions, f_star, selected_features
            )
            if termination_condition != TerminationCondition.infeasible and (
                best_model is None or tmp_model.z.value < best_model.z.value
            ):
                best_model = tmp_model
            else:
                raise ValueError("Unable to find a partition for f* = %s" % f_star)

        # Solve for exact number of regions >= 2 when f* was not defined beforehand
        if self.beta is None or self.exact_number_regions == 2:
            self.final_model = best_model
        elif self.exact_number_regions is not None:
            number_regions = self.exact_number_regions
            tmp_model, termination_condition = self.run_model(
                X, y, number_regions, f_star, selected_features
            )
            self.final_model = tmp_model
        # Otherwise, keep increasing the number of regions until best model is achieved
        else:
            f_star = best_model.fStar
            previousZ = float("inf")
            currentZ = best_model.z.value

            while currentZ < previousZ * (1 - self.beta):
                number_regions += 1
                if verbose:
                    print(
                        "\n%s\n########## R = %d f* = %s ###########"
                        % (self, number_regions, f_star)
                    )

                tmp_model, termination_condition = self.run_model(
                    X, y, number_regions, f_star, selected_features
                )
                if termination_condition == TerminationCondition.infeasible:
                    break
                previousZ, currentZ = currentZ, tmp_model.z.value
                if currentZ < previousZ * (1 - self.beta):
                    best_model = tmp_model

            self.final_model = best_model
        return self

    def get_model_info(self):
        coefficients = pd.DataFrame()

        featureNames = self.final_model.get_selected_features()
        cols = np.array("region")
        cols = np.append(cols, featureNames)
        cols = np.append(cols, "B")

        for region in range(self.final_model.number_regions):
            values = {"region": region}
            for f in featureNames:
                values.update({f: self.final_model.W[region, f].value})
            values.update({"B": self.final_model.B[region].value})
            coefficients = coefficients.append(
                pd.DataFrame(data=values, index=np.arange(1), columns=cols),
                ignore_index=True,
            )

        if self.final_model.number_regions == 1:
            breakpointsDF = pd.DataFrame(
                data={"fStar": None, "breakpoints": None}, index=np.arange(1)
            )
        else:
            breakpointsDF = pd.DataFrame(
                {
                    "fStar": self.final_model.fStar,
                    "breakpoints": self.final_model.get_breakpoints(),
                }
            )

        return coefficients, breakpointsDF



import numpy as np
import pandas as pd

from pyomo.core.base import (
    ConcreteModel,
    Var,
    Set,
    NonNegativeReals,
    Binary,
    Constraint,
    Objective,
    minimize,
    Reals,
)


class OplraPyomoModel(ConcreteModel):
    """A Pyomo concrete model that defines OPLRA mathematical programming model

    Attributes:
        data         (DataFrame): data frame containing samples and features
        target         (object) : a list or pandas Series of the target variables of regression
        number_regions         (int) : if specified, runs Oplra to the exact number of regions
        lam             (float) : lambda regularisation parameter
        epsilon         (float) : MIP epsilon parameter (usually set at 0.01)
        nMax              (int) : if specified, add constraints that limit the maximum number of features to be selected
        fStar             (str) : partition feature f* to be used in the model. Only valid for noRegions >= 2
        selectedFeatures (list) : subset of features to be used in the MIP model. Usually defined after running OPLRA
                                    for noRegions == 1 and selecting those features with nonzero coefficients
        sampleWeight     (list) : Weights assigned to the samples
        isSimpleLinear   (bool) : if True, sets all features to selected. (Probably will be deprecated)
    """

    def __init__(
        self,
        data,
        target,
        number_regions,
        lam,
        epsilon=0.01,
        f_star=None,
        selected_features=None,
    ):
        super().__init__(
            name="OplraRegression %d regions (f*=%s)" % (number_regions, f_star)
        )
        self.number_regions = number_regions

        # Data and parameters
        self.data = data
        self.target = target
        self.lam = lam
        self.U1 = 1.5
        self.U2 = sum(self.target)
        self.epsilon = epsilon

        # Indices of the model
        self.s = Set(initialize=data.index.tolist())
        self.r = Set(initialize=range(self.number_regions))
        if self.number_regions == 1 or selected_features is None:
            self.f = Set(initialize=self.data.columns, ordered=True)
        else:
            self.f = Set(initialize=selected_features, ordered=True)

        # Variables common to both FEATURE_SELECTION and PIECEWISE models
        self.W = Var(self.r, self.f, initialize=0)

        self.absW = Var(self.r, self.f, domain=NonNegativeReals, initialize=0)
        self.B = Var(self.r, initialize=0)
        self.Pred = Var(self.r, self.s, initialize=0)
        self.D = Var(self.s, domain=NonNegativeReals, initialize=0)
        self.F = Var(self.r, self.s, domain=Binary, initialize=0)
        self.z = Var(domain=NonNegativeReals, initialize=0)
        self.mae = Var(domain=NonNegativeReals, initialize=0)
        self.reg = Var(domain=NonNegativeReals, initialize=0)

        if self.number_regions > 1 and f_star is None:
            raise ValueError("f_star is invalid.")

        self.fStar = f_star

        # Specific variables and constraints according to model (number of regions)
        if self.number_regions == 1:
            self.sf = Var(self.f, domain=Binary, initialize=0)
            # Old equations used for explicit feature selection
            # self.number_features    = Constraint(rule=OplraPyomoModel.number_features_rule)
            # self.upper_bound_coeff  = Constraint(self.r, self.f, rule=OplraPyomoModel.upper_bound_coeff_rule)
            # self.lower_bound_coeff  = Constraint(self.r, self.f, rule=OplraPyomoModel.lower_bound_coeff_rule)
        else:
            self.isSimpleLinear = False
            self.fStar = self.fStar
            self.X = Var(self.r, bounds=(0.0, 1.0), initialize=0)
            self.rr = Set(within=self.r)
            self.breakpoint_order = Constraint(
                self.r, rule=OplraPyomoModel.breakpoint_order_rule
            )
            self.region_divider1 = Constraint(
                self.r, self.s, rule=OplraPyomoModel.region_divider1_rule
            )
            self.region_divider2 = Constraint(
                self.r, self.s, rule=OplraPyomoModel.region_divider2_rule
            )

        # Common constraints
        self.sample_to_region = Constraint(
            self.s, rule=OplraPyomoModel.sample_to_region_rule
        )
        self.prediction = Constraint(
            self.r, self.s, rule=OplraPyomoModel.prediction_rule
        )
        self.abs_error1 = Constraint(
            self.r, self.s, rule=OplraPyomoModel.abs_error1_rule
        )
        self.abs_error2 = Constraint(
            self.r, self.s, rule=OplraPyomoModel.abs_error2_rule
        )
        self.abs_coeff1 = Constraint(
            self.r, self.f, rule=OplraPyomoModel.abs_coeff1_rule
        )
        self.abs_coeff2 = Constraint(
            self.r, self.f, rule=OplraPyomoModel.abs_coeff2_rule
        )
        self.mae_eqn = Constraint(rule=OplraPyomoModel.mae_rule)
        self.reg_eqn = Constraint(rule=OplraPyomoModel.regularisation_rule)
        self.z_eqn = Constraint(rule=OplraPyomoModel.z_rule)
        self.obj = Objective(rule=OplraPyomoModel.objective_rule, sense=minimize)

    # ---- UTIL FUNCTIONS ---- #
    def get_selected_features(self):
        fs = []
        for r in self.r:
            for f in self.f:
                if self.W[r, f].value != 0 and f not in fs:
                    fs.append(f)
        return fs

    def get_breakpoints(self):
        breakpoints = [self.X[r].value for r in self.r]
        del breakpoints[-1]
        return breakpoints

    def get_regions(self, f_star_series):
        samples = range(len(f_star_series))

        if type(f_star_series) == pd.Series:
            f_star_series.index = samples
        breakpoints = self.get_breakpoints()
        regions = list(self.r)
        lo_r = {r: breakpoints[r - 1] if r != 0 else 0.0 for r in regions}
        up_r = {r: breakpoints[r] if r != len(regions) - 1 else 1 for r in regions}
        ts_distance = [
            [
                max(0, f_star_series[s] - up_r[r], lo_r[r] - f_star_series[s])
                for r in regions
            ]
            for s in samples
        ]
        ts_region = [ts_distance[s].index(min(ts_distance[s])) for s in samples]
        return np.array(ts_region)

    def get_coefficients(self):
        # TODO (improvement): use SparseCoefMixin sparsify
        return np.vstack(
            [self.W[(r, f)].value if f in self.f else 0 for f in self.data.columns]
            for r in self.r
        )

    def get_intercepts(self):
        return np.array([B.value for B in list(self.B.values())])

    # ----- EQUATIONS ---- #
    def breakpoint_order_rule(model, r):
        if r == model.number_regions - 1:
            return Constraint.Skip
        else:
            return model.X[r] + model.epsilon <= model.X[r + 1]

    def region_divider1_rule(model, r, s):
        if r == model.number_regions - 1:
            return Constraint.Skip
        else:
            return (
                model.X[r] - model.epsilon + model.U1 * (1 - model.F[r, s])
            ) >= model.data.loc[s, model.fStar]

    def region_divider2_rule(model, r, s):
        if r == 0:
            return Constraint.Skip
        else:
            return (
                model.X[r - 1] + model.epsilon - model.U1 * (1 - model.F[r, s])
            ) <= model.data.loc[s, model.fStar]

    def sample_to_region_rule(model, s):
        return sum(model.F[rr, s] for rr in model.r) == 1

    def prediction_rule(model, r, s):
        return model.Pred[r, s] == (
            sum(model.data.loc[s, ff] * model.W[r, ff] for ff in model.f) + model.B[r]
        )

    def abs_error1_rule(model, r, s):
        return model.D[s] >= (
            model.target[s] - model.Pred[r, s] - model.U2 * (1 - model.F[r, s])
        )

    def abs_error2_rule(model, r, s):
        return model.D[s] >= (
            model.Pred[r, s] - model.target[s] - model.U2 * (1 - model.F[r, s])
        )

    def abs_coeff1_rule(model, r, f):
        return model.absW[r, f] >= model.W[r, f]

    def abs_coeff2_rule(model, r, f):
        return model.absW[r, f] >= -model.W[r, f]

    def mae_rule(model):
        return model.mae == sum(model.D[ss] for ss in model.s) / len(model.s)

    def regularisation_rule(model):
        return model.reg == model.lam * sum(
            model.absW[r, f] for r in model.r for f in model.f
        )

    def z_rule(model):
        return model.z == model.mae + model.reg

    def objective_rule(model):
        return model.z

    # ----- FEATURE SELECTION EQUATIONS ---- #
    def number_features_rule(model):
        return sum(model.sf[f] for f in model.f) <= model.n_max

    def upper_bound_coeff_rule(model, r, f):
        return model.W[r, f] <= model.U2 * model.sf[f]

    def lower_bound_coeff_rule(model, r, f):
        return model.W[r, f] >= -model.U2 * model.sf[f]



from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.tree import DecisionTreeRegressor, export_graphviz

from .models import OplraPyomoModel
from .solvers import get_solver_definition


class BaseOplraEstimator(BaseEstimator, RegressorMixin):
    """Superclass of estimators used in this package
    """

    def __init__(
        self, algorithm_name, algorithm_version, lam, epsilon, beta, solver_name
    ):
        self.algorithm_name = algorithm_name
        self.algorithm_version = algorithm_version
        self.lam = lam
        self.epsilon = epsilon
        self.beta = beta
        self.solver_name = solver_name
        self.solver_def = get_solver_definition(solver_name)

    def get_model_info(self):
        raise NotImplementedError("Function should be implemented in algorithm's class")

    def run_model(self, X, y, number_regions, f_star=None, selected_features=None):
        pyomo_model = build_piecewise_model(
            X, y, self.lam, self.epsilon, number_regions, f_star, selected_features
        )
        termination_condition = self.solver_def.solve(pyomo_model, verbose=False)
        return pyomo_model, termination_condition


def build_piecewise_model(
    data,
    target,
    lam,
    epsilon,
    number_regions,
    f_star=None,
    selected_features=None,
    n_max=None,
):
    """
        Helper function to create a OplraPyomoModel
        :param data:
        :param target:
        :param number_regions:
        :param f_star:
        :param selected_features:
        :param n_max:
        :return:
    """

    return OplraPyomoModel(
        data,
        target,
        number_regions,
        lam=lam,
        epsilon=epsilon,
        f_star=f_star,
        selected_features=selected_features,
    )



from setuptools import setup, find_packages

setup(
    name="oplra_reg",
    description="Mixed Integer Piecewise Linear Regression with Regularisation",
    long_description=(
        "OPLRAreg is a regression technique based on mathematical programming that splits "
        "data into separate regions and identifies independent linear equations for each region."
    ),
    version="0.3",
    py_modules=["oplrareg"],
    author="Jonathan Cardoso Silva",
    author_email="jonathan.car.silva@gmail.com",
    url="https://github.com/KISysBio/OPLRAreg",
    url_download="https://github.com/KISysBio/OPLRAreg/archive/v0.2.1.tar.gz",
    keywords=[""],
    install_requires=[
        "numpy >= 1.13",
        "scipy >= 0.19.1",
        "pandas >= 0.20.3",
        "scikit-learn >= 0.19.0",
        "pyomo >= 5.3",
    ],
    entry_points={"console_scripts": ["oplrareg = oplrareg.__main__:main"]},
    packages=find_packages(),
    python_requires=">=3",
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)
